const { inspect } = require('util');
const { PublisherId, PublicSigningKey } = require('../peer/identity');
const pb = require('../protobuf');
const serialize = require('../metadata/serialize');
const { b58MultihashForBuffer, stringifyNestedBuffers, setUnion } = require('../common/util');

class Statement {

  static fromProtobuf(msg) {
    const body = StatementBody.fromProtobuf(msg.body);

    const { id, publisher, namespace, timestamp, signature } = msg;
    if (signature == null) {
      return new UnsignedStatement({ id, publisher, namespace, timestamp, body });
    }

    return new SignedStatement({ id, publisher, namespace, timestamp, body, signature });
  }

  static fromBytes(bytes) {
    return Statement.fromProtobuf(pb.stmt.Statement.decode(bytes));
  }

  toProtobuf() {
    const { id, publisher, namespace, timestamp, body } = this;
    const msg = { id, publisher, namespace, timestamp: timestamp, body: body.toProtobuf() };
    if (this instanceof SignedStatement) {
      msg.signature = this.signature;
    }
    return msg;
  }

  toBytes() {
    return pb.stmt.Statement.encode(this.toProtobuf());
  }

  toJSON() {
    return stringifyNestedBuffers(this.toProtobuf());
  }

  inspect(_depth, opts) {
    opts = Object.assign({}, opts, { depth: null });
    const { id, publisher, namespace, timestamp, body } = this;
    const output = { id, publisher, namespace, timestamp, body };
    if (this instanceof SignedStatement) {
      output.signature = this.signature;
    }
    return inspect(stringifyNestedBuffers(output), opts);
  }

  get objectIds() {
    return this.body.objectIds;
  }

  get refSet() {
    return this.body.refSet;
  }

  get source() {
    if (!(this.body instanceof EnvelopeStatementBody)) {
      return this.publisher;
    }

    if (this.body.statements.length < 1) {
      return this.publisher;
    }

    return this.body.statements[0].source;
  }

  expandObjects(source) {
    const body = this.body.expandObjects(source);
    const { id, publisher, namespace, timestamp } = this;
    if (this instanceof SignedStatement) {
      return new SignedStatement({ id, publisher, namespace, timestamp, body, signature: this.signature });
    }
    return new UnsignedStatement({ id, publisher, namespace, timestamp, body });
  }

  asUnsignedStatement() {
    const { id, namespace, publisher, timestamp, body } = this;
    return new UnsignedStatement({ id, namespace, publisher, timestamp, body });
  }
}

class UnsignedStatement extends Statement {
  constructor(stmt) {
    super();
    this.id = stmt.id;
    this.publisher = stmt.publisher;
    this.namespace = stmt.namespace;
    this.timestamp = stmt.timestamp;
    this.body = stmt.body;
  }

  sign(publisherId) {
    return Promise.resolve().then(() => {
      if (publisherId.id58 !== this.publisher) {
        throw new Error(`Cannot sign statement, publisher id of signer does not match statement publisher`);
      }

      return this.calculateSignature(publisherId).then(signature => {
        const { id, namespace, publisher, timestamp, body } = this;
        return new SignedStatement({ id, namespace, publisher, timestamp, body, signature });
      });
    });
  }

  calculateSignature(signer) {
    return signer.sign(this.toBytes());
  }
}

class SignedStatement extends Statement {

  constructor(stmt) {
    super();
    this.id = stmt.id;
    this.publisher = stmt.publisher;
    this.namespace = stmt.namespace;
    this.timestamp = stmt.timestamp;
    this.body = stmt.body;
    if (Buffer.isBuffer(stmt.signature)) {
      this.signature = stmt.signature;
    } else {
      this.signature = Buffer.from(stmt.signature, 'base64');
    }
  }

  static create(publisherId, namespace, statementBody, counter = 0, timestampGenerator = Date.now) {
    const body = statementBody instanceof StatementBody ? statementBody : StatementBody.fromProtobuf(statementBody);
    const timestamp = timestampGenerator();
    const statementId = [publisherId.id58, timestamp.toString(), counter.toString()].join(':');
    const stmt = new UnsignedStatement({
      id: statementId,
      publisher: publisherId.id58,
      namespace,
      timestamp,
      body
    });
    return stmt.sign(publisherId);
  }

  static createSimple(publisherId, namespace, statementBody, counter = 0, timestampGenerator = Date.now) {
    let body;
    if (typeof statementBody.object === 'object') {
      body = new ExpandedSimpleStatementBody(statementBody);
    } else {
      body = new SimpleStatementBody(statementBody);
    }
    return SignedStatement.create(publisherId, namespace, body, counter, timestampGenerator);
  }

  static fromProtobuf(msg) {
    const body = StatementBody.fromProtobuf(msg.body);

    const { id, publisher, namespace, timestamp, signature } = msg;
    if (signature == null) {
      throw new Error('SignedStatement.fromProtobuf() requires a non-null signature. ' + 'Use Statement.fromProtobuf() or UnsignedStatement.fromProtobuf() if your message is unsigned.');
    }
    return new SignedStatement({ id, publisher, namespace, timestamp, body, signature });
  }

  verifySignature(publicKey, keyCache) {
    return Promise.resolve().then(() => {
      keyCache = keyCache || new Map();
      publicKey = publicKey || keyCache.get(this.publisher) || PublicSigningKey.fromB58String(this.publisher);
      keyCache.set(this.publisher, publicKey);

      const msg = this.toProtobuf();
      msg.signature = null;
      const bytes = pb.stmt.Statement.encode(msg);
      return publicKey.verify(bytes, this.signature);
    });
  }
}

class StatementBody {
  constructor() {
    if (this.constructor === StatementBody) {
      throw new Error("StatementBody class is abstract, don't instantiate directly");
    }
  }

  static fromProtobuf(msg) {
    if (msg.simple != null) {
      return SimpleStatementBody.fromProtobuf(msg.simple);
    } else if (msg.compound != null) {
      return CompoundStatementBody.fromProtobuf(msg.compound);
    } else if (msg.envelope != null) {
      return EnvelopeStatementBody.fromProtobuf(msg.envelope);
    }

    throw new Error('Unsupported statement body ' + JSON.stringify(msg));
  }

  toProtobuf() {
    throw new Error('StatementBody.toProtobuf() is not implemented - subclasses must implement');
  }

  expandObjects(source) {
    return this;
  }

  get refSet() {
    return new Set();
  }

  get objectIds() {
    return [];
  }
}

class SimpleStatementBody extends StatementBody {

  constructor(contents) {
    super();
    this.objectRef = contents.object;
    this.refs = contents.refs || [];
    this.deps = contents.deps || [];
    this.tags = contents.tags || [];
  }

  static fromProtobuf(msg) {
    return new SimpleStatementBody(msg);
  }

  toProtobuf() {
    return { simple: this.toSimpleProtobuf() };
  }

  toSimpleProtobuf() {
    const { objectRef, refs, deps, tags } = this;
    return { object: objectRef, refs, deps, tags };
  }

  expandObjects(source) {
    const object = source.get(this.objectRef);
    if (object == null) {
      // FIXME: should we just return the un-expanded body instead?
      throw new Error(`No object matching ref ${this.objectRef} available, cannot expand`);
    }

    return new ExpandedSimpleStatementBody({ object, refs: this.refs, deps: this.deps, tags: this.tags });
  }

  get objectIds() {
    return [this.objectRef];
  }

  get refSet() {
    return new Set(this.refs);
  }

  inspect(_depth, _opts) {
    return this.toSimpleProtobuf();
  }
}

class CompoundStatementBody extends StatementBody {

  constructor(simpleBodies) {
    super();
    this.simpleBodies = simpleBodies;
  }

  static fromProtobuf(msg) {
    return new CompoundStatementBody(msg.body.map(b => SimpleStatementBody.fromProtobuf(b)));
  }

  toProtobuf() {
    return {
      compound: {
        body: this.simpleBodies.map(b => b.toSimpleProtobuf())
      }
    };
  }

  expandObjects(source) {
    const expanded = this.simpleBodies.map(b => b.expandObjects(source));
    return new CompoundStatementBody(expanded);
  }

  inspect(_depth, _opts) {
    return this.simpleBodies;
  }

  get objectIds() {
    return [].concat(...this.simpleBodies.map(b => b.objectIds));
  }

  get refSet() {
    return setUnion(...this.simpleBodies.map(b => b.refSet));
  }
}

class EnvelopeStatementBody extends StatementBody {

  constructor(statements) {
    super();
    this.statements = statements;
  }

  static fromProtobuf(msg) {
    return new EnvelopeStatementBody(msg.body.map(stmt => SignedStatement.fromProtobuf(stmt)));
  }

  toProtobuf() {
    return {
      envelope: {
        body: this.statements.map(stmt => stmt.toProtobuf())
      }
    };
  }

  expandObjects(source) {
    const expanded = this.statements.map(stmt => stmt.expandObjects(source));
    return new EnvelopeStatementBody(expanded);
  }

  inspect(_depth, _opts) {
    return this.statements;
  }

  get objectIds() {
    return [].concat(...this.statements.map(stmt => stmt.objectIds));
  }

  get refSet() {
    return setUnion(...this.statements.map(stmt => stmt.refSet));
  }
}

// Expanded statement bodies include the data that a SimpleStatementBody's `object` hash links to
class ExpandedSimpleStatementBody extends SimpleStatementBody {

  constructor(contents) {
    const object = contents.object;
    const bytes = serialize.encode(object);
    const ref = b58MultihashForBuffer(bytes);
    super({ object: ref, refs: contents.refs, deps: contents.deps, tags: contents.tags });
    this.object = object;
  }

  toJSON() {
    const { object, refs, deps, tags } = this;
    return { object, refs, deps, tags };
  }

  inspect(_depth, opts) {
    opts = Object.assign({}, opts, { depth: null });
    return inspect(this.toJSON(), opts);
  }
}

module.exports = {
  Statement,
  UnsignedStatement,
  SignedStatement,
  StatementBody,
  SimpleStatementBody,
  CompoundStatementBody,
  EnvelopeStatementBody,
  ExpandedSimpleStatementBody
};